package mvd.jester;

import java.lang.reflect.Constructor;
import java.time.LocalTime;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import com.google.common.collect.HashBasedTable;
import com.google.common.collect.Table;
import mvd.jester.model.SystemSetup;
import mvd.jester.priority.PriorityManager;
import mvd.jester.simulator.AbstractSimulator;
import mvd.jester.tests.AbstractTest;
import mvd.jester.utils.Logger;

/**
 * TestEnvironment
 */
public class TestEnvironment {

    private final long numberOfTaskSets;
    private final SystemSetup systemSetup;
    private final SystemSetup.Builder builder;
    private final Set<Constructor<? extends AbstractTest>> abstractTests;
    private final Set<Constructor<? extends AbstractSimulator>> abstractSimulators;
    private final Set<PriorityManager> schedulingAlgorithms;

    public TestEnvironment(SystemSetup.Builder builder, long numberOfTaskSets) {
        this.numberOfTaskSets = numberOfTaskSets;
        this.abstractTests = new HashSet<>();
        this.abstractSimulators = new HashSet<>();
        this.schedulingAlgorithms = new HashSet<>();

        this.builder = builder;
        this.systemSetup = builder.build();
    }

    public TestEnvironment registerSchedulingAlgorithm(PriorityManager priorityManager) {
        schedulingAlgorithms.add(priorityManager);

        return this;
    }

    public TestEnvironment registerTest(Class<? extends AbstractTest> abstractTest) {
        try {
            abstractTests.add(abstractTest.getConstructor(SystemSetup.class));
        } catch (Exception e) {
            System.out.println("Missing constructor for abstract test!");
        }

        return this;
    }

    public TestEnvironment registerSimulator(Class<? extends AbstractSimulator> abstractSimulator) {
        try {
            abstractSimulators.add(abstractSimulator.getConstructor(SystemSetup.class));
        } catch (Exception e) {
            System.out.println("Missing constructor for abstract simulator!");
        }

        return this;
    }

    public TestEnvironment registerTestPair(Class<? extends AbstractTest> abstractTest,
            Class<? extends AbstractSimulator> abstractSimulator) {
        try {
            abstractTests.add(abstractTest.getConstructor(SystemSetup.class));
        } catch (Exception e) {
            System.out.println("Missing constructor for abstract test!");
        }
        try {
            abstractSimulators.add(abstractSimulator.getConstructor(SystemSetup.class));
        } catch (Exception e) {
            System.out.println("Missing constructor for abstract simulator!");
        }

        return this;
    }

    public void runTests() {
        Set<AbstractTest> abstractTestInstances = new HashSet<>();
        Set<AbstractSimulator> abstractSimulatorInstances = new HashSet<>();

        Map<Long, Long> totalNumberOfTasks = new HashMap<>();

        Map<Long, Table<PriorityManager, AbstractTest, Long>> testResults = new HashMap<>();
        Map<Long, Table<PriorityManager, AbstractSimulator, Long>> simulatorResults =
                new HashMap<>();

        for (Constructor<? extends AbstractTest> c : abstractTests) {
            try {
                abstractTestInstances.add(c.newInstance(this.systemSetup));
            } catch (Exception e) {
                System.out.println("Could not instantiate object of AbstractTest!");
            }
        }

        for (Constructor<? extends AbstractSimulator> c : abstractSimulators) {
            try {
                abstractSimulatorInstances.add(c.newInstance(this.systemSetup));
            } catch (Exception e) {
                System.out.println("Could not instantiate object of AbstractSimulator!");
            }
        }

        long checkedTasksets = 0;

        while (checkedTasksets < numberOfTaskSets) {
            builder.rebuild(this.systemSetup);

            double utilization = this.systemSetup.getUtilization();

            while (utilization <= this.systemSetup.getNumberOfProcessors()
                    && checkedTasksets < numberOfTaskSets) {
                checkedTasksets++;

                if (checkedTasksets % 100 == 0) {
                    System.out.println(checkedTasksets + " bereits geschafft");
                }

                long roundedUtilization = (long) (utilization * 10);

                totalNumberOfTasks.compute(roundedUtilization, (k, v) -> v == null ? 1 : v + 1);

                for (PriorityManager priorityManager : schedulingAlgorithms) {
                    for (AbstractTest test : abstractTestInstances) {
                        Table<PriorityManager, AbstractTest, Long> table = testResults
                                .computeIfAbsent(roundedUtilization, k -> HashBasedTable.create());
                        if (priorityManager.hasTest(test)
                                && test.runSchedulabilityCheck(priorityManager)) {
                            Long val = table.get(priorityManager, test);
                            if (val == null) {
                                table.put(priorityManager, test, (long) 1);
                            } else {
                                table.put(priorityManager, test, val + 1);
                            }
                        }
                    }

                    for (AbstractSimulator simulator : abstractSimulatorInstances) {
                        Table<PriorityManager, AbstractSimulator, Long> table = simulatorResults
                                .computeIfAbsent(roundedUtilization, k -> HashBasedTable.create());
                        if (priorityManager.hasSimulator(simulator)
                                && simulator.runSimulation(priorityManager)) {
                            Long val = table.get(priorityManager, simulator);
                            if (val == null) {
                                table.put(priorityManager, simulator, (long) 1);
                            } else {
                                table.put(priorityManager, simulator, val + 1);
                            }
                        }
                    }

                }
                builder.addTask(systemSetup);

                utilization = this.systemSetup.getUtilization();
            }
        }

        logTestResults(testResults, totalNumberOfTasks);
        logSimulationResults(simulatorResults, totalNumberOfTasks);
    }

    private void logTestResults(Map<Long, Table<PriorityManager, AbstractTest, Long>> results,
            Map<Long, Long> totalNumberOfTasks) {
        LocalTime date = LocalTime.now();
        Logger log = new Logger("./results/results_" + systemSetup.getNumberOfProcessors() + "_"
                + date.getHour() + ":" + date.getMinute() + ".txt");
        String firstLine = new String("Utilization");

        if (!results.isEmpty()) {
            Long first = results.keySet().iterator().next();

            for (PriorityManager p : results.get(first).rowKeySet()) {
                for (AbstractTest t : results.get(first).columnKeySet()) {
                    firstLine = firstLine + "\t" + t.getName() + "_" + p.getName();
                }
            }

            log.log(firstLine);

            for (Long util : results.keySet()) {
                String line = String.valueOf((double) util / 10);
                Table<PriorityManager, AbstractTest, Long> tests = results.get(util);

                for (PriorityManager p : tests.rowKeySet()) {
                    for (AbstractTest t : tests.columnKeySet()) {
                        if (tests.contains(p, t)) {
                            line += "\t" + tests.get(p, t);
                        } else {
                            line += "\t" + "0";
                        }
                    }
                }
                log.log(line);
            }
        }

        log.finalize();
    }

    private void logSimulationResults(
            Map<Long, Table<PriorityManager, AbstractSimulator, Long>> results,
            Map<Long, Long> totalNumberOfTasks) {
        LocalTime date = LocalTime.now();
        Logger log = new Logger("./results/results_sim_" + systemSetup.getNumberOfProcessors() + "_"
                + date.getHour() + ":" + date.getMinute() + ".txt");
        String firstLine = new String("Utilization\tTotal");

        if (!results.isEmpty()) {
            Long first = results.keySet().iterator().next();

            for (PriorityManager p : results.get(first).rowKeySet()) {
                for (AbstractSimulator s : results.get(first).columnKeySet()) {
                    firstLine = firstLine + "\t" + s.getName() + "_" + p.getName();
                }
            }

            log.log(firstLine);

            for (Long util : totalNumberOfTasks.keySet()) {
                String line = String.valueOf((double) util / 10);
                line += "\t" + totalNumberOfTasks.get(util);
                Table<PriorityManager, AbstractSimulator, Long> tests = results.get(util);

                for (PriorityManager p : tests.rowKeySet()) {
                    for (AbstractSimulator s : tests.columnKeySet()) {
                        if (tests.contains(p, s)) {
                            line += "\t" + tests.get(p, s);
                        } else {
                            line += "\t" + "0";
                        }
                    }
                }
                log.log(line);
            }
        }

        log.finalize();
    }
}
